@@ -22,6 +22,7 @@
 #include "gtd-edit-pane.h"
 #include "gtd-empty-list-widget.h"
 #include "gtd-task-list-view.h"
+#include "gtd-task-list-view-model.h"
 #include "gtd-manager.h"
 #include "gtd-markdown-renderer.h"
 #include "gtd-new-task-row.h"
@@ -57,28 +58,26 @@
  *
  * gtd_task_list_view_set_model (view, model);
  *
- * // Hide the '+ New task' row
- * gtd_task_list_view_set_show_new_task_row (view, FALSE);
- *
  * // Date which tasks will be automatically assigned
  * gtd_task_list_view_set_default_date (view, now);
  * ]|
  *
  */
 
-typedef struct
+struct _GtdTaskListView
 {
+  GtkBox                  parent;
+
   GtdEmptyListWidget    *empty_list_widget;
   GtkListBox            *listbox;
   GtkStack              *main_stack;
-  GtkListBoxRow         *new_task_row;
   GtkWidget             *scrolled_window;
-  GtkStack              *stack;
 
   /* internal */
   gboolean               can_toggle;
   gboolean               show_due_date;
   gboolean               show_list_name;
+  GtdTaskListViewModel  *view_model;
   GListModel            *model;
   GDateTime             *default_date;
 
@@ -111,34 +110,8 @@ typedef struct
   GtdTaskRow             *active_row;
   GtkSizeGroup           *due_date_sizegroup;
   GtkSizeGroup           *tasklist_name_sizegroup;
-  GtdTaskListSelectorBehavior task_list_selector_behavior;
-} GtdTaskListViewPrivate;
-
-struct _GtdTaskListView
-{
-  GtkBox                  parent;
-
-  /*<private>*/
-  GtdTaskListViewPrivate *priv;
 };
 
-typedef enum
-{
-  GTD_IDLE_STATE_STARTED,
-  GTD_IDLE_STATE_LOADING,
-  GTD_IDLE_STATE_COMPLETE,
-  GTD_IDLE_STATE_FINISHED,
-} GtdIdleState;
-
-typedef struct
-{
-  GtdTaskListView      *self;
-  GtdIdleState          state;
-  GPtrArray            *added;
-  GPtrArray            *removed;
-  guint32               current_item;
-} GtdIdleData;
-
 #define DND_SCROLL_OFFSET            24 //px
 #define TASK_REMOVED_NOTIFICATION_ID "task-removed-id"
 
@@ -168,7 +141,7 @@ static void          on_task_row_exited_cb                       (GtdTaskListVie
 static gboolean      scroll_to_bottom_cb                         (gpointer            data);
 
 
-G_DEFINE_TYPE_WITH_PRIVATE (GtdTaskListView, gtd_task_list_view, GTK_TYPE_BOX)
+G_DEFINE_TYPE (GtdTaskListView, gtd_task_list_view, GTK_TYPE_BOX)
 
 static const GActionEntry gtd_task_list_view_entries[] = {
   { "clear-completed-tasks", on_clear_completed_tasks_activated_cb },
@@ -184,7 +157,6 @@ enum {
   PROP_0,
   PROP_SHOW_LIST_NAME,
   PROP_SHOW_DUE_DATE,
-  PROP_SHOW_NEW_TASK_ROW,
   LAST_PROP
 };
 
@@ -196,22 +168,25 @@ enum {
 static inline GtdTaskRow*
 task_row_from_row (GtkListBoxRow *row)
 {
-  return GTD_TASK_ROW (gtk_list_box_row_get_child (row));
+  GtkWidget *child = gtk_list_box_row_get_child (row);
+
+  if (!GTD_IS_TASK_ROW (child))
+    return NULL;
+
+  return GTD_TASK_ROW (child);
 }
 
 static void
 set_active_row (GtdTaskListView *self,
                 GtdTaskRow      *row)
 {
-  GtdTaskListViewPrivate *priv = gtd_task_list_view_get_instance_private (self);
-
-  if (priv->active_row == row)
+  if (self->active_row == row)
     return;
 
-  if (priv->active_row)
-    gtd_task_row_set_active (priv->active_row, FALSE);
+  if (self->active_row)
+    gtd_task_row_set_active (self->active_row, FALSE);
 
-  priv->active_row = row;
+  self->active_row = row;
 
   if (row)
     {
@@ -223,41 +198,36 @@ set_active_row (GtdTaskListView *self,
 static void
 schedule_scroll_to_bottom (GtdTaskListView *self)
 {
-  GtdTaskListViewPrivate *priv = gtd_task_list_view_get_instance_private (self);
-
-  if (priv->scroll_to_bottom_handler_id > 0)
+  if (self->scroll_to_bottom_handler_id > 0)
     return;
 
-  priv->scroll_to_bottom_handler_id = g_timeout_add (250, scroll_to_bottom_cb, self);
+  self->scroll_to_bottom_handler_id = g_timeout_add (250, scroll_to_bottom_cb, self);
 }
 
 static void
 update_empty_state (GtdTaskListView *self)
 {
-  GtdTaskListViewPrivate *priv = gtd_task_list_view_get_instance_private (self);
   gboolean show_empty_list_widget;
   gboolean is_empty;
 
   g_assert (GTD_IS_TASK_LIST_VIEW (self));
 
-  if (!priv->model)
+  if (!self->model)
     return;
 
-  is_empty = g_list_model_get_n_items (priv->model) == 0;
-  gtd_empty_list_widget_set_is_empty (priv->empty_list_widget, is_empty);
+  is_empty = g_list_model_get_n_items (self->model) == 0;
+  gtd_empty_list_widget_set_is_empty (self->empty_list_widget, is_empty);
 
-  show_empty_list_widget = !GTD_IS_TASK_LIST (priv->model) &&
-                           (is_empty || priv->n_incomplete_tasks == 0);
-  gtk_stack_set_visible_child_name (priv->main_stack,
+  show_empty_list_widget = !GTD_IS_TASK_LIST (self->model) &&
+                           (is_empty || self->n_incomplete_tasks == 0);
+  gtk_stack_set_visible_child_name (self->main_stack,
                                     show_empty_list_widget ? "empty-list" : "task-list");
 }
 
 static void
 update_incomplete_tasks_model (GtdTaskListView *self)
 {
-  GtdTaskListViewPrivate *priv = gtd_task_list_view_get_instance_private (self);
-
-  if (!priv->incomplete_tasks_model)
+  if (!self->incomplete_tasks_model)
     {
       g_autoptr (GtkFilterListModel) filter_model = NULL;
       GtkCustomFilter *filter;
@@ -266,14 +236,14 @@ update_incomplete_tasks_model (GtdTaskListView *self)
       filter_model = gtk_filter_list_model_new (NULL, GTK_FILTER (filter));
       gtk_filter_list_model_set_incremental (filter_model, TRUE);
 
-      priv->incomplete_tasks_model = G_LIST_MODEL (g_steal_pointer (&filter_model));
+      self->incomplete_tasks_model = G_LIST_MODEL (g_steal_pointer (&filter_model));
     }
 
-  gtk_filter_list_model_set_model (GTK_FILTER_LIST_MODEL (priv->incomplete_tasks_model),
-                                   priv->model);
-  priv->n_incomplete_tasks = g_list_model_get_n_items (priv->incomplete_tasks_model);
+  gtk_filter_list_model_set_model (GTK_FILTER_LIST_MODEL (self->incomplete_tasks_model),
+                                   self->model);
+  self->n_incomplete_tasks = g_list_model_get_n_items (self->incomplete_tasks_model);
 
-  g_signal_connect (priv->incomplete_tasks_model,
+  g_signal_connect (self->incomplete_tasks_model,
                     "items-changed",
                     G_CALLBACK (on_incomplete_tasks_items_changed_cb),
                     self);
@@ -299,10 +269,8 @@ on_incomplete_tasks_items_changed_cb (GListModel      *model,
                                       guint            n_added,
                                       GtdTaskListView *self)
 {
-  GtdTaskListViewPrivate *priv = gtd_task_list_view_get_instance_private (self);
-
-  priv->n_incomplete_tasks -= n_removed;
-  priv->n_incomplete_tasks += n_added;
+  self->n_incomplete_tasks -= n_removed;
+  self->n_incomplete_tasks += n_added;
 
   update_empty_state (self);
 }
@@ -311,39 +279,61 @@ static void
 on_empty_list_widget_add_tasks_cb (GtdEmptyListWidget *empty_list_widget,
                                    GtdTaskListView    *self)
 {
-  GtdTaskListViewPrivate *priv = gtd_task_list_view_get_instance_private (self);
+  gtk_stack_set_visible_child_name (self->main_stack, "task-list");
+}
 
-  gtk_stack_set_visible_child_name (priv->main_stack, "task-list");
+static void
+on_new_task_row_entered_cb (GtdTaskListView *self,
+                            GtdNewTaskRow   *row)
+{
+  set_active_row (self, NULL);
+}
+
+static void
+on_new_task_row_exited_cb (GtdTaskListView *self,
+                           GtdNewTaskRow   *row)
+{
 }
 
 static GtkWidget*
 create_row_for_task_cb (gpointer item,
                         gpointer user_data)
 {
-  GtdTaskListViewPrivate *priv;
   GtdTaskListView *self;
   GtkWidget *listbox_row;
   GtkWidget *row;
 
   self = GTD_TASK_LIST_VIEW (user_data);
-  priv = gtd_task_list_view_get_instance_private (self);
-
-  row = gtd_task_row_new (item, priv->renderer);
-
-  gtd_task_row_set_list_name_visible (GTD_TASK_ROW (row), priv->show_list_name);
-  gtd_task_row_set_due_date_visible (GTD_TASK_ROW (row), priv->show_due_date);
-
-  g_signal_connect_swapped (row, "enter", G_CALLBACK (on_task_row_entered_cb), self);
-  g_signal_connect_swapped (row, "exit", G_CALLBACK (on_task_row_exited_cb), self);
-
-  g_signal_connect (row, "remove-task", G_CALLBACK (on_remove_task_row_cb), self);
 
   listbox_row = gtk_list_box_row_new ();
+
+  if (GTD_IS_TASK (item))
+    {
+      row = gtd_task_row_new (item, self->renderer);
+
+      gtd_task_row_set_list_name_visible (GTD_TASK_ROW (row), self->show_list_name);
+      gtd_task_row_set_due_date_visible (GTD_TASK_ROW (row), self->show_due_date);
+
+      g_signal_connect_swapped (row, "enter", G_CALLBACK (on_task_row_entered_cb), self);
+      g_signal_connect_swapped (row, "exit", G_CALLBACK (on_task_row_exited_cb), self);
+
+      g_signal_connect (row, "remove-task", G_CALLBACK (on_remove_task_row_cb), self);
+    }
+  else
+    {
+      row = gtd_new_task_row_new ();
+
+      g_signal_connect_swapped (row, "enter", G_CALLBACK (on_new_task_row_entered_cb), self);
+      g_signal_connect_swapped (row, "exit", G_CALLBACK (on_new_task_row_exited_cb), self);
+
+      gtk_list_box_row_set_activatable (GTK_LIST_BOX_ROW (listbox_row), FALSE);
+    }
+
   gtk_list_box_row_set_child (GTK_LIST_BOX_ROW (listbox_row), row);
 
   g_object_bind_property (row, "visible", listbox_row, "visible", G_BINDING_BIDIRECTIONAL);
 
-  g_hash_table_insert (priv->task_to_row, item, row);
+  g_hash_table_insert (self->task_to_row, item, row);
 
   return listbox_row;
 }
@@ -351,18 +341,17 @@ create_row_for_task_cb (gpointer item,
 static gboolean
 scroll_to_bottom_cb (gpointer data)
 {
-  GtdTaskListViewPrivate *priv;
+  GtdTaskListView *self = GTD_TASK_LIST_VIEW (data);
   GtkWidget *widget;
   GtkRoot *root;
 
-  priv = gtd_task_list_view_get_instance_private (data);
-  widget = GTK_WIDGET (data);
+  widget = GTK_WIDGET (self);
   root = gtk_widget_get_root (widget);
 
   if (!root)
     return G_SOURCE_CONTINUE;
 
-  priv->scroll_to_bottom_handler_id = 0;
+  self->scroll_to_bottom_handler_id = 0;
 
   /*
    * Only focus the new task row if the current list is visible,
@@ -373,10 +362,12 @@ scroll_to_bottom_cb (gpointer data)
       gtk_widget_get_mapped (widget) &&
       !gtk_widget_is_ancestor (gtk_window_get_focus (GTK_WINDOW (root)), widget))
     {
+      GtkWidget *new_task_row;
       gboolean ignored;
 
-      gtk_widget_grab_focus (GTK_WIDGET (priv->new_task_row));
-      g_signal_emit_by_name (priv->scrolled_window, "scroll-child", GTK_SCROLL_END, FALSE, &ignored);
+      new_task_row = gtk_widget_get_last_child (GTK_WIDGET (self->listbox));
+      gtk_widget_grab_focus (new_task_row);
+      g_signal_emit_by_name (self->scrolled_window, "scroll-child", GTK_SCROLL_END, FALSE, &ignored);
     }
 
   return G_SOURCE_REMOVE;
@@ -405,7 +396,7 @@ on_clear_completed_tasks_activated_cb (GSimpleAction *simple,
   guint i;
 
   self = GTD_TASK_LIST_VIEW (user_data);
-  model = self->priv->model;
+  model = self->model;
 
   for (i = 0; i < g_list_model_get_n_items (model); i++)
     {
@@ -499,20 +490,6 @@ on_remove_task_row_cb (GtdTaskRow      *row,
   set_active_row (self, NULL);
 }
 
-static void
-on_new_task_row_entered_cb (GtdTaskListView *self,
-                            GtdNewTaskRow   *row)
-{
-  set_active_row (self, NULL);
-}
-
-static void
-on_new_task_row_exited_cb (GtdTaskListView *self,
-                           GtdNewTaskRow   *row)
-{
-
-}
-
 static void
 on_task_row_entered_cb (GtdTaskListView *self,
                         GtdTaskRow      *row)
@@ -524,9 +501,7 @@ static void
 on_task_row_exited_cb (GtdTaskListView *self,
                        GtdTaskRow      *row)
 {
-  GtdTaskListViewPrivate *priv = self->priv;
-
-  if (row == priv->active_row)
+  if (row == self->active_row)
     set_active_row (self, NULL);
 }
 
@@ -541,6 +516,9 @@ on_listbox_row_activated_cb (GtkListBox      *listbox,
 
   task_row = task_row_from_row (row);
 
+  if (!task_row)
+    GTD_RETURN ();
+
   /* Toggle the row */
   if (gtd_task_row_get_active (task_row))
     set_active_row (self, NULL);
@@ -558,24 +536,27 @@ on_listbox_row_activated_cb (GtkListBox      *listbox,
 static void
 internal_header_func (GtkListBoxRow   *row,
                       GtkListBoxRow   *before,
-                      GtdTaskListView *view)
+                      GtdTaskListView *self)
 {
   GtkWidget *header;
   GtdTask *row_task;
   GtdTask *before_task;
 
-  if (!view->priv->header_func)
+  if (!self->header_func)
     return;
 
   row_task = before_task = NULL;
 
+  if (!task_row_from_row (row))
+    return;
+
   if (row)
     row_task = gtd_task_row_get_task (task_row_from_row (row));
 
   if (before)
       before_task = gtd_task_row_get_task (task_row_from_row (before));
 
-  header = view->priv->header_func (row_task, before_task, view->priv->header_user_data);
+  header = self->header_func (row_task, before_task, self->header_user_data);
 
   if (header)
     {
@@ -596,18 +577,16 @@ internal_header_func (GtkListBoxRow   *row,
 static void
 clear_drag_highlights (GtdTaskListView *self)
 {
-  GtdTaskListViewPrivate *priv = gtd_task_list_view_get_instance_private (self);
-
-  if (priv->drag_highlight.top)
+  if (self->drag_highlight.top)
     {
-      gtk_widget_remove_css_class (priv->drag_highlight.top, "top-highlight");
-      priv->drag_highlight.top = NULL;
+      gtk_widget_remove_css_class (self->drag_highlight.top, "top-highlight");
+      self->drag_highlight.top = NULL;
     }
 
-  if (priv->drag_highlight.bottom)
+  if (self->drag_highlight.bottom)
     {
-      gtk_widget_remove_css_class (priv->drag_highlight.bottom, "bottom-highlight");
-      priv->drag_highlight.bottom = NULL;
+      gtk_widget_remove_css_class (self->drag_highlight.bottom, "bottom-highlight");
+      self->drag_highlight.bottom = NULL;
     }
 }
 
@@ -615,14 +594,12 @@ static void
 update_row_drag_highlight (GtdTaskListView *self,
                            gdouble          y)
 {
-  GtdTaskListViewPrivate *priv;
   GtkAllocation row_allocation;
   GtkListBoxRow *hovered_row;
   GtkListBoxRow *bottom_highlight;
   GtkListBoxRow *top_highlight;
 
-  priv = gtd_task_list_view_get_instance_private (self);
-  hovered_row = gtk_list_box_get_row_at_y (priv->listbox, y);
+  hovered_row = gtk_list_box_get_row_at_y (self->listbox, y);
   bottom_highlight = NULL;
   top_highlight = NULL;
 
@@ -671,32 +648,35 @@ update_row_drag_highlight (GtdTaskListView *self,
         }
     }
 
+  /* Don't add drag highlights to the new task row */
+  if (top_highlight && !GTD_IS_TASK_ROW (task_row_from_row (top_highlight)))
+    top_highlight = NULL;
+  if (bottom_highlight && !GTD_IS_TASK_ROW (task_row_from_row (bottom_highlight)))
+    bottom_highlight = NULL;
+
   /* Unhighlight previously highlighted rows */
   clear_drag_highlights (self);
 
-  priv->drag_highlight.top = GTK_WIDGET (top_highlight);
-  priv->drag_highlight.bottom = GTK_WIDGET (bottom_highlight);
+  self->drag_highlight.top = GTK_WIDGET (top_highlight);
+  self->drag_highlight.bottom = GTK_WIDGET (bottom_highlight);
 
   /* Highlight new rows */
-  if (priv->drag_highlight.top)
-    gtk_widget_add_css_class (priv->drag_highlight.top, "top-highlight");
-  if (priv->drag_highlight.bottom)
-    gtk_widget_add_css_class (priv->drag_highlight.bottom, "bottom-highlight");
+  if (self->drag_highlight.top)
+    gtk_widget_add_css_class (self->drag_highlight.top, "top-highlight");
+  if (self->drag_highlight.bottom)
+    gtk_widget_add_css_class (self->drag_highlight.bottom, "bottom-highlight");
 }
 
 static GtkListBoxRow*
 get_drop_row_at_y (GtdTaskListView *self,
                    gdouble          y)
 {
-  GtdTaskListViewPrivate *priv;
   GtkAllocation row_allocation;
   GtkListBoxRow *hovered_row;
   GtkListBoxRow *task_row;
   GtkListBoxRow *drop_row;
 
-  priv = gtd_task_list_view_get_instance_private (self);
-
-  hovered_row = gtk_list_box_get_row_at_y (priv->listbox, y);
+  hovered_row = gtk_list_box_get_row_at_y (self->listbox, y);
 
   /* Small optimization when hovering the first row */
   if (gtk_list_box_row_get_index (hovered_row) == 0)
@@ -733,19 +713,18 @@ get_drop_row_at_y (GtdTaskListView *self,
       drop_row = task_row;
     }
 
-  return drop_row ? drop_row : NULL;
+  return task_row_from_row (drop_row) ? drop_row : NULL;
 }
 
 static inline gboolean
 scroll_to_dnd (gpointer user_data)
 {
-  GtdTaskListViewPrivate *priv;
+  GtdTaskListView *self = GTD_TASK_LIST_VIEW (user_data);
   GtkAdjustment *vadjustment;
   gint value;
 
-  priv = gtd_task_list_view_get_instance_private (user_data);
-  vadjustment = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (priv->scrolled_window));
-  value = gtk_adjustment_get_value (vadjustment) + (priv->scroll_up ? -6 : 6);
+  vadjustment = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (self->scrolled_window));
+  value = gtk_adjustment_get_value (vadjustment) + (self->scroll_up ? -6 : 6);
 
   gtk_adjustment_set_value (vadjustment,
                             CLAMP (value, 0, gtk_adjustment_get_upper (vadjustment)));
@@ -758,45 +737,44 @@ check_dnd_scroll (GtdTaskListView *self,
                   gboolean         should_cancel,
                   gdouble          y)
 {
-  GtdTaskListViewPrivate *priv = gtd_task_list_view_get_instance_private (self);
   gdouble current_y, height;
 
   if (should_cancel)
     {
-      if (priv->scroll_timeout_id > 0)
+      if (self->scroll_timeout_id > 0)
         {
-          g_source_remove (priv->scroll_timeout_id);
-          priv->scroll_timeout_id = 0;
+          g_source_remove (self->scroll_timeout_id);
+          self->scroll_timeout_id = 0;
         }
 
       return;
     }
 
-  height = gtk_widget_get_allocated_height (priv->scrolled_window);
-  gtk_widget_translate_coordinates (GTK_WIDGET (priv->listbox),
-                                    priv->scrolled_window,
+  height = gtk_widget_get_allocated_height (self->scrolled_window);
+  gtk_widget_translate_coordinates (GTK_WIDGET (self->listbox),
+                                    self->scrolled_window,
                                     0, y,
                                     NULL, &current_y);
 
   if (current_y < DND_SCROLL_OFFSET || current_y > height - DND_SCROLL_OFFSET)
     {
-      if (priv->scroll_timeout_id > 0)
+      if (self->scroll_timeout_id > 0)
         return;
 
       /* Start the autoscroll */
-      priv->scroll_up = current_y < DND_SCROLL_OFFSET;
-      priv->scroll_timeout_id = g_timeout_add (25,
+      self->scroll_up = current_y < DND_SCROLL_OFFSET;
+      self->scroll_timeout_id = g_timeout_add (25,
                                                scroll_to_dnd,
                                                self);
     }
   else
     {
-      if (priv->scroll_timeout_id == 0)
+      if (self->scroll_timeout_id == 0)
         return;
 
       /* Cancel the autoscroll */
-      g_source_remove (priv->scroll_timeout_id);
-      priv->scroll_timeout_id = 0;
+      g_source_remove (self->scroll_timeout_id);
+      self->scroll_timeout_id = 0;
     }
 }
 
@@ -860,7 +838,6 @@ on_drop_target_drag_drop_cb (GtkDropTarget   *drop_target,
                              gdouble          y,
                              GtdTaskListView *self)
 {
-  GtdTaskListViewPrivate *priv;
   GtkListBoxRow *drop_row;
   GtdTaskRow *hovered_row;
   GtkWidget *row;
@@ -873,7 +850,6 @@ on_drop_target_drag_drop_cb (GtkDropTarget   *drop_target,
 
   GTD_ENTRY;
 
-  priv = gtd_task_list_view_get_instance_private (self);
   drop = gtk_drop_target_get_drop (drop_target);
   drag = gdk_drop_get_drag (drop);
 
@@ -892,10 +868,16 @@ on_drop_target_drag_drop_cb (GtkDropTarget   *drop_target,
    * When the drag operation began, the source row was hidden. Now is the time
    * to show it again.
    */
-  row = g_hash_table_lookup (priv->task_to_row, source_task);
+  row = g_hash_table_lookup (self->task_to_row, source_task);
   gtk_widget_show (row);
 
   drop_row = get_drop_row_at_y (self, y);
+  if (!drop_row)
+    {
+      check_dnd_scroll (self, TRUE, -1);
+      GTD_RETURN (FALSE);
+    }
+
   hovered_row = task_row_from_row (drop_row);
   hovered_task = gtd_task_row_get_task (hovered_row);
   new_position = gtd_task_get_position (hovered_task);
@@ -905,7 +887,7 @@ on_drop_target_drag_drop_cb (GtkDropTarget   *drop_target,
 
   if (new_position != current_position)
     {
-      gtd_task_list_move_task_to_position (GTD_TASK_LIST (priv->model),
+      gtd_task_list_move_task_to_position (GTD_TASK_LIST (self->model),
                                            source_task,
                                            new_position);
     }
@@ -923,14 +905,14 @@ on_drop_target_drag_drop_cb (GtkDropTarget   *drop_target,
 static void
 gtd_task_list_view_finalize (GObject *object)
 {
-  GtdTaskListViewPrivate *priv = GTD_TASK_LIST_VIEW (object)->priv;
+  GtdTaskListView *self = GTD_TASK_LIST_VIEW (object);
 
-  g_clear_handle_id (&priv->scroll_to_bottom_handler_id, g_source_remove);
-  g_clear_pointer (&priv->task_to_row, g_hash_table_destroy);
-  g_clear_pointer (&priv->default_date, g_date_time_unref);
-  g_clear_object (&priv->incomplete_tasks_model);
-  g_clear_object (&priv->renderer);
-  g_clear_object (&priv->model);
+  g_clear_handle_id (&self->scroll_to_bottom_handler_id, g_source_remove);
+  g_clear_pointer (&self->task_to_row, g_hash_table_destroy);
+  g_clear_pointer (&self->default_date, g_date_time_unref);
+  g_clear_object (&self->incomplete_tasks_model);
+  g_clear_object (&self->renderer);
+  g_clear_object (&self->model);
 
   G_OBJECT_CLASS (gtd_task_list_view_parent_class)->finalize (object);
 }
@@ -946,15 +928,11 @@ gtd_task_list_view_get_property (GObject    *object,
   switch (prop_id)
     {
     case PROP_SHOW_DUE_DATE:
-      g_value_set_boolean (value, self->priv->show_due_date);
+      g_value_set_boolean (value, self->show_due_date);
       break;
 
     case PROP_SHOW_LIST_NAME:
-      g_value_set_boolean (value, self->priv->show_list_name);
-      break;
-
-    case PROP_SHOW_NEW_TASK_ROW:
-      g_value_set_boolean (value, gtk_widget_get_visible (GTK_WIDGET (self->priv->new_task_row)));
+      g_value_set_boolean (value, self->show_list_name);
       break;
 
     default:
@@ -980,10 +958,6 @@ gtd_task_list_view_set_property (GObject      *object,
       gtd_task_list_view_set_show_list_name (self, g_value_get_boolean (value));
       break;
 
-    case PROP_SHOW_NEW_TASK_ROW:
-      gtd_task_list_view_set_show_new_task_row (self, g_value_get_boolean (value));
-      break;
-
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
@@ -993,13 +967,12 @@ static void
 gtd_task_list_view_constructed (GObject *object)
 {
   GtdTaskListView *self = GTD_TASK_LIST_VIEW (object);
-
   G_OBJECT_CLASS (gtd_task_list_view_parent_class)->constructed (object);
 
   /* action_group */
-  self->priv->action_group = G_ACTION_GROUP (g_simple_action_group_new ());
+  self->action_group = G_ACTION_GROUP (g_simple_action_group_new ());
 
-  g_action_map_add_action_entries (G_ACTION_MAP (self->priv->action_group),
+  g_action_map_add_action_entries (G_ACTION_MAP (self->action_group),
                                    gtd_task_list_view_entries,
                                    G_N_ELEMENTS (gtd_task_list_view_entries),
                                    object);
@@ -1014,7 +987,6 @@ static void
 gtd_task_list_view_map (GtkWidget *widget)
 {
   GtdTaskListView *self = GTD_TASK_LIST_VIEW (widget);
-  GtdTaskListViewPrivate *priv = gtd_task_list_view_get_instance_private (self);
   GtkRoot *root;
 
 
@@ -1028,7 +1000,7 @@ gtd_task_list_view_map (GtkWidget *widget)
   gtk_widget_insert_action_group (GTK_WIDGET (root), "list", NULL);
 
   /* Add this instance's action group */
-  gtk_widget_insert_action_group (GTK_WIDGET (root), "list", priv->action_group);
+  gtk_widget_insert_action_group (GTK_WIDGET (root), "list", self->action_group);
 }
 
 static void
@@ -1049,20 +1021,6 @@ gtd_task_list_view_class_init (GtdTaskListViewClass *klass)
   g_type_ensure (GTD_TYPE_TASK_ROW);
   g_type_ensure (GTD_TYPE_EMPTY_LIST_WIDGET);
 
-  /**
-   * GtdTaskListView::show-new-task-row:
-   *
-   * Whether the list shows the "New Task" row or not.
-   */
-  g_object_class_install_property (
-        object_class,
-        PROP_SHOW_NEW_TASK_ROW,
-        g_param_spec_boolean ("show-new-task-row",
-                              "Whether it shows the New Task row",
-                              "Whether the list shows the New Task row, or not",
-                              TRUE,
-                              G_PARAM_READWRITE));
-
   /**
    * GtdTaskListView::show-list-name:
    *
@@ -1093,14 +1051,12 @@ gtd_task_list_view_class_init (GtdTaskListViewClass *klass)
 
   gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/todo/ui/gtd-task-list-view.ui");
 
-  gtk_widget_class_bind_template_child_private (widget_class, GtdTaskListView, due_date_sizegroup);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdTaskListView, empty_list_widget);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdTaskListView, listbox);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdTaskListView, main_stack);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdTaskListView, new_task_row);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdTaskListView, tasklist_name_sizegroup);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdTaskListView, scrolled_window);
-  gtk_widget_class_bind_template_child_private (widget_class, GtdTaskListView, stack);
+  gtk_widget_class_bind_template_child (widget_class, GtdTaskListView, due_date_sizegroup);
+  gtk_widget_class_bind_template_child (widget_class, GtdTaskListView, empty_list_widget);
+  gtk_widget_class_bind_template_child (widget_class, GtdTaskListView, listbox);
+  gtk_widget_class_bind_template_child (widget_class, GtdTaskListView, main_stack);
+  gtk_widget_class_bind_template_child (widget_class, GtdTaskListView, tasklist_name_sizegroup);
+  gtk_widget_class_bind_template_child (widget_class, GtdTaskListView, scrolled_window);
 
   gtk_widget_class_bind_template_callback (widget_class, on_empty_list_widget_add_tasks_cb);
   gtk_widget_class_bind_template_callback (widget_class, on_listbox_row_activated_cb);
@@ -1115,19 +1071,13 @@ gtd_task_list_view_class_init (GtdTaskListViewClass *klass)
 static void
 gtd_task_list_view_init (GtdTaskListView *self)
 {
-  GtdTaskListViewPrivate *priv;
   GtkDropTarget *target;
 
-  priv = gtd_task_list_view_get_instance_private (self);
+  self->task_to_row = g_hash_table_new (NULL, NULL);
 
-  self->priv = priv;
-
-  priv->task_list_selector_behavior = GTD_TASK_LIST_SELECTOR_BEHAVIOR_AUTOMATIC;
-  priv->task_to_row = g_hash_table_new (NULL, NULL);
-
-  priv->can_toggle = TRUE;
-  priv->show_due_date = TRUE;
-  priv->show_due_date = TRUE;
+  self->can_toggle = TRUE;
+  self->show_due_date = TRUE;
+  self->show_due_date = TRUE;
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
@@ -1138,9 +1088,16 @@ gtd_task_list_view_init (GtdTaskListView *self)
   g_signal_connect (target, "leave", G_CALLBACK (on_drop_target_drag_leave_cb), self);
   g_signal_connect (target, "motion", G_CALLBACK (on_drop_target_drag_motion_cb), self);
 
-  gtk_widget_add_controller (GTK_WIDGET (priv->listbox), GTK_EVENT_CONTROLLER (target));
+  gtk_widget_add_controller (GTK_WIDGET (self->listbox), GTK_EVENT_CONTROLLER (target));
 
-  priv->renderer = gtd_markdown_renderer_new ();
+  self->renderer = gtd_markdown_renderer_new ();
+
+  self->view_model = gtd_task_list_view_model_new ();
+  gtk_list_box_bind_model (self->listbox,
+                           G_LIST_MODEL (self->view_model),
+                           create_row_for_task_cb,
+                           self,
+                           NULL);
 }
 
 /**
@@ -1156,38 +1113,6 @@ gtd_task_list_view_new (void)
   return g_object_new (GTD_TYPE_TASK_LIST_VIEW, NULL);
 }
 
-/**
- * gtd_task_list_view_get_show_new_task_row:
- * @view: a #GtdTaskListView
- *
- * Gets whether @view shows the new task row or not.
- *
- * Returns: %TRUE if @view is shows the new task row, %FALSE otherwise
- */
-gboolean
-gtd_task_list_view_get_show_new_task_row (GtdTaskListView *self)
-{
-  g_return_val_if_fail (GTD_IS_TASK_LIST_VIEW (self), FALSE);
-
-  return gtk_widget_get_visible (GTK_WIDGET (self->priv->new_task_row));
-}
-
-/**
- * gtd_task_list_view_set_show_new_task_row:
- * @view: a #GtdTaskListView
- *
- * Sets the #GtdTaskListView:show-new-task-mode property of @view.
- */
-void
-gtd_task_list_view_set_show_new_task_row (GtdTaskListView *view,
-                                          gboolean         show_new_task_row)
-{
-  g_return_if_fail (GTD_IS_TASK_LIST_VIEW (view));
-
-  gtk_widget_set_visible (GTK_WIDGET (view->priv->new_task_row), show_new_task_row);
-  g_object_notify (G_OBJECT (view), "show-new-task-row");
-}
-
 /**
  * gtd_task_list_view_get_model:
  * @view: a #GtdTaskListView
@@ -1202,7 +1127,7 @@ gtd_task_list_view_get_model (GtdTaskListView *view)
 {
   g_return_val_if_fail (GTD_IS_TASK_LIST_VIEW (view), NULL);
 
-  return view->priv->model;
+  return view->model;
 }
 
 /**
@@ -1217,28 +1142,18 @@ void
 gtd_task_list_view_set_model (GtdTaskListView *view,
                               GListModel      *model)
 {
-  GtdTaskListViewPrivate *priv;
-
   g_return_if_fail (GTD_IS_TASK_LIST_VIEW (view));
   g_return_if_fail (G_IS_LIST_MODEL (model));
 
-  priv = gtd_task_list_view_get_instance_private (view);
-
-  if (!g_set_object (&priv->model, model))
+  if (view->model == model)
     return;
 
-  gtk_list_box_bind_model (priv->listbox,
-                           model,
-                           create_row_for_task_cb,
-                           view,
-                           NULL);
+  view->model = model;
 
+  gtd_task_list_view_model_set_model (view->view_model, model);
   schedule_scroll_to_bottom (view);
   update_incomplete_tasks_model (view);
   update_empty_state (view);
-
-  if (priv->task_list_selector_behavior == GTD_TASK_LIST_SELECTOR_BEHAVIOR_AUTOMATIC)
-    gtd_new_task_row_set_show_list_selector (GTD_NEW_TASK_ROW (priv->new_task_row), !GTD_IS_TASK_LIST (model));
 }
 
 /**
@@ -1254,7 +1169,7 @@ gtd_task_list_view_get_show_list_name (GtdTaskListView *view)
 {
   g_return_val_if_fail (GTD_IS_TASK_LIST_VIEW (view), FALSE);
 
-  return view->priv->show_list_name;
+  return view->show_list_name;
 }
 
 /**
@@ -1270,13 +1185,13 @@ gtd_task_list_view_set_show_list_name (GtdTaskListView *view,
 {
   g_return_if_fail (GTD_IS_TASK_LIST_VIEW (view));
 
-  if (view->priv->show_list_name != show_list_name)
+  if (view->show_list_name != show_list_name)
     {
       GtkWidget *child;
 
-      view->priv->show_list_name = show_list_name;
+      view->show_list_name = show_list_name;
 
-      for (child = gtk_widget_get_first_child (GTK_WIDGET (view->priv->listbox));
+      for (child = gtk_widget_get_first_child (GTK_WIDGET (view->listbox));
            child;
            child = gtk_widget_get_next_sibling (child))
         {
@@ -1306,7 +1221,7 @@ gtd_task_list_view_get_show_due_date (GtdTaskListView *self)
 {
   g_return_val_if_fail (GTD_IS_TASK_LIST_VIEW (self), FALSE);
 
-  return self->priv->show_due_date;
+  return self->show_due_date;
 }
 
 /**
@@ -1320,19 +1235,16 @@ void
 gtd_task_list_view_set_show_due_date (GtdTaskListView *self,
                                       gboolean         show_due_date)
 {
-  GtdTaskListViewPrivate *priv;
   GtkWidget *child;
 
   g_return_if_fail (GTD_IS_TASK_LIST_VIEW (self));
 
-  priv = gtd_task_list_view_get_instance_private (self);
-
-  if (priv->show_due_date == show_due_date)
+  if (self->show_due_date == show_due_date)
     return;
 
-  priv->show_due_date = show_due_date;
+  self->show_due_date = show_due_date;
 
-  for (child = gtk_widget_get_first_child (GTK_WIDGET (priv->listbox));
+  for (child = gtk_widget_get_first_child (GTK_WIDGET (self->listbox));
        child;
        child = gtk_widget_get_next_sibling (child))
     {
@@ -1363,28 +1275,24 @@ gtd_task_list_view_set_header_func (GtdTaskListView           *view,
                                     GtdTaskListViewHeaderFunc  func,
                                     gpointer                   user_data)
 {
-  GtdTaskListViewPrivate *priv;
-
   g_return_if_fail (GTD_IS_TASK_LIST_VIEW (view));
 
-  priv = view->priv;
-
   if (func)
     {
-      priv->header_func = func;
-      priv->header_user_data = user_data;
+      view->header_func = func;
+      view->header_user_data = user_data;
 
-      gtk_list_box_set_header_func (priv->listbox,
+      gtk_list_box_set_header_func (view->listbox,
                                     (GtkListBoxUpdateHeaderFunc) internal_header_func,
                                     view,
                                     NULL);
     }
   else
     {
-      priv->header_func = NULL;
-      priv->header_user_data = NULL;
+      view->header_func = NULL;
+      view->header_user_data = NULL;
 
-      gtk_list_box_set_header_func (priv->listbox,
+      gtk_list_box_set_header_func (view->listbox,
                                     NULL,
                                     NULL,
                                     NULL);
@@ -1402,13 +1310,9 @@ gtd_task_list_view_set_header_func (GtdTaskListView           *view,
 GDateTime*
 gtd_task_list_view_get_default_date (GtdTaskListView *self)
 {
-  GtdTaskListViewPrivate *priv;
-
   g_return_val_if_fail (GTD_IS_TASK_LIST_VIEW (self), NULL);
 
-  priv = gtd_task_list_view_get_instance_private (self);
-
-  return priv->default_date;
+  return self->default_date;
 }
 
 /**
@@ -1422,62 +1326,11 @@ void
 gtd_task_list_view_set_default_date   (GtdTaskListView *self,
                                        GDateTime       *default_date)
 {
-  GtdTaskListViewPrivate *priv;
-
   g_return_if_fail (GTD_IS_TASK_LIST_VIEW (self));
 
-  priv = gtd_task_list_view_get_instance_private (self);
-
-  if (priv->default_date == default_date)
+  if (self->default_date == default_date)
     return;
 
-  g_clear_pointer (&priv->default_date, g_date_time_unref);
-  priv->default_date = default_date ? g_date_time_ref (default_date) : NULL;
-}
-
-GtdTaskListSelectorBehavior
-gtd_task_list_view_get_task_list_selector_behavior (GtdTaskListView *self)
-{
-  GtdTaskListViewPrivate *priv;
-
-  g_return_val_if_fail (GTD_IS_TASK_LIST_VIEW (self), -1);
-
-  priv = gtd_task_list_view_get_instance_private (self);
-
-  return priv->task_list_selector_behavior;
-}
-
-void
-gtd_task_list_view_set_task_list_selector_behavior (GtdTaskListView             *self,
-                                                    GtdTaskListSelectorBehavior  behavior)
-{
-  GtdTaskListViewPrivate *priv;
-
-  g_return_if_fail (GTD_IS_TASK_LIST_VIEW (self));
-
-  priv = gtd_task_list_view_get_instance_private (self);
-
-  if (priv->task_list_selector_behavior == behavior)
-    return;
-
-  priv->task_list_selector_behavior = behavior;
-
-  switch (behavior)
-    {
-    case GTD_TASK_LIST_SELECTOR_BEHAVIOR_AUTOMATIC:
-      if (priv->model)
-        {
-          gtd_new_task_row_set_show_list_selector (GTD_NEW_TASK_ROW (priv->new_task_row),
-                                                   !GTD_IS_TASK_LIST (priv->model));
-        }
-      break;
-
-    case GTD_TASK_LIST_SELECTOR_BEHAVIOR_ALWAYS_SHOW:
-      gtd_new_task_row_set_show_list_selector (GTD_NEW_TASK_ROW (priv->new_task_row), TRUE);
-      break;
-
-    case GTD_TASK_LIST_SELECTOR_BEHAVIOR_ALWAYS_HIDE:
-      gtd_new_task_row_set_show_list_selector (GTD_NEW_TASK_ROW (priv->new_task_row), FALSE);
-      break;
-    }
+  g_clear_pointer (&self->default_date, g_date_time_unref);
+  self->default_date = default_date ? g_date_time_ref (default_date) : NULL;
 }
